
import unittest
from time import perf_counter_ns

import numpy as np
from scipy.constants import c, pi

from lwrad import get_lw_spectrum


def dummy_thomson():
    import h5py

    l0 = 0.8e-6
    with h5py.File("tests/test.h5", "r") as f:
        x = f['x'][:, 0] / 2/pi * l0
        y = f['y'][:, 0] / 2/pi * l0
        z = f['z'][:, 0] / 2/pi * l0
        ux = f['px'][:, 0]
        uy = f['py'][:, 0]
        uz = f['pz'][:, 0]
        t = np.arange(len(ux)) * f.attrs['dt'] / 2/pi * l0/c

    return x, y, z, ux, uy, uz, t


class TestSpeed(unittest.TestCase):
    
    def setUp(self) -> None:
        l0 = 8e-7
        omega0 = 2*pi*c / l0
        self.args = dummy_thomson()
        self.args_multiple = [np.stack([arg]*1000, axis=1) for arg in self.args[:-1]]
        self.args_multiple += [self.args[-1]]
        self.n = [1, 0, 0]
        nomega = 256
        gamma0 = 5.0

        self.omega_axis = np.linspace(0, 3, nomega) * 4*gamma0**2*omega0

    def test_cpu(self):
        get_lw_spectrum(*self.args, self.n, self.omega_axis, backend=None)
        tic = perf_counter_ns()
        for _ in range(100):
            get_lw_spectrum(*self.args, self.n, self.omega_axis, backend=None)
        dt = perf_counter_ns() - tic
        print(f"get_lw_spectrum on cpu: {dt/100*1e-6:.0f} ms/call")


    def test_cpu_multiple(self):
        get_lw_spectrum(*self.args_multiple, self.n, self.omega_axis)
        tic = perf_counter_ns()
        for _ in range(100):
            get_lw_spectrum(*self.args_multiple, self.n, self.omega_axis)
        dt = perf_counter_ns() - tic
        print(f"get_lw_spectrum 2d on cpu: {dt/100*1e-6:.0f} ms/call")


    def test_cuda(self):
        try:
            from lwrad.cuda import get_lw_spectrum_cuda
        except ImportError as e:
            print("CUDA not available")
        get_lw_spectrum_cuda(*self.args, self.n, self.omega_axis)
        tic = perf_counter_ns()
        for _ in range(100):
            get_lw_spectrum_cuda(*self.args, self.n, self.omega_axis)
        dt = perf_counter_ns() - tic
        print(f"get_lw_spectrum_cuda: {dt/100*1e-6:.0f} ms/call")

    def test_cuda_multiple(self):
        try:
            from lwrad.cuda import get_lw_spectrum_2d_cuda
        except ImportError as e:
            print("CUDA not available")
        
        get_lw_spectrum_2d_cuda(*self.args_multiple, self.n, self.omega_axis)
        tic = perf_counter_ns()
        for _ in range(100):
            get_lw_spectrum(*self.args_multiple, self.n, self.omega_axis)
        dt = perf_counter_ns() - tic
        print(f"get_lw_spectrum_2d_cuda: {dt/100*1e-6:.0f} ms/call")


import numpy as np
from scipy.constants import pi, m_e, e, c, alpha, hbar, epsilon_0, mu_0
from numba import njit, prange
from warnings import warn

def get_lw_RE(x, y, z, ux, uy, uz, t, n):
    '''
    从坐标和动量计算推迟时间和电场，实际为计算R*E，忽略速度项

    参考doi.org/10.5281/zenodo.843510 2.32、2.36、2.39式

    x, y, z : ndarray
        坐标向量
    ux, uy, uz : ndarray
        归一化动量，u = p/mc
    t : ndarray
        时间矢量
    n : tuple | list | ndarray
        长度为1的方向矢量
    '''

    if np.isnan(x).any():
        warn("NaN values detected in input vector. NaNs will be discarded.")

    nx = n[0]
    ny = n[1]
    nz = n[2]
    n_norm = np.sqrt(nx**2 + ny**2 + nz**2)
    if n_norm != 1:
        nx /= n_norm
        ny /= n_norm
        nz /= n_norm

    return calc_lw(x, y, z, ux, uy, uz, t, nx, ny, nz)


def _calc_lw(x, y, z, ux, uy, uz, t, nx, ny, nz):
    nt = len(t)
    t_ret = np.zeros(nt-2)
    REx = np.zeros(nt-2)
    REy = np.zeros(nt-2)
    REz = np.zeros(nt-2)
    for it in prange(1, nt-1):

        t_ret[it-1] = t[it] - (nx*x[it] + ny*y[it] + nz*z[it]) / c

        betax, betay, betaz = _calc_beta(ux[it], uy[it], uz[it])
        betax_prev, betay_prev, betaz_prev = _calc_beta(ux[it-1], uy[it-1], uz[it-1])
        betax_next, betay_next, betaz_next = _calc_beta(ux[it+1], uy[it+1], uz[it+1])

        dt = t[it+1] - t[it-1]
        # 加速度，假设首尾加速度为0
        ax = (betax_next - betax_prev) / dt
        ay = (betay_next - betay_prev) / dt
        az = (betaz_next - betaz_prev) / dt

        n_dot_a = nx*ax + ny*ay + nz*az
        n_dot_beta = nx*betax + ny*betay + nz*betaz
        factor = 1 / (1 - n_dot_beta)**3
        factor *= e / (4*pi*epsilon_0*c)

        REx[it-1] = (n_dot_a*(nx - betax) + (n_dot_beta - 1) * ax) * factor
        REy[it-1] = (n_dot_a*(ny - betay) + (n_dot_beta - 1) * ay) * factor
        REz[it-1] = (n_dot_a*(nz - betaz) + (n_dot_beta - 1) * az) * factor
    return t_ret, REx, REy, REz

calc_lw_mt = njit(_calc_lw, parallel=True)
calc_lw = njit(_calc_lw)


@njit(inline="always")
def _calc_beta(ux, uy, uz):
    inv_gamma = 1 / np.sqrt(ux**2 + uy**2 + uz**2 + 1)
    betax = ux * inv_gamma
    betay = uy * inv_gamma
    betaz = uz * inv_gamma
    return betax, betay, betaz

@njit(parallel=True)
def get_RE_spectrum_mt(RE, t_ret, omega_axis):
    '''
    计算lw谱，不使用FFT直接积分。慢但是方便。

    RE : ndarray
        推迟势电场矢量REx, REy or REz
    t_ret : ndarray
        推迟时间矢量
    omega_axis : ndarray
        频谱的频率轴
    '''
    
    nomega = len(omega_axis)
    RE_ft = np.zeros(nomega, dtype=np.complex128)
    norm = 1 / np.sqrt(c*mu_0) / np.sqrt(2*pi)
    
    # definition of Fourier transformation
    for i in prange(nomega):
        w = omega_axis[i]
        # trapzoid integral
        RE_ft[i] = np.trapz(RE * np.exp(1j*w*t_ret), t_ret) * norm

    # normalize
    return RE_ft

@njit
def get_RE_spectrum_st(RE, t_ret, omega_axis):
    '''
    计算lw谱，不使用FFT直接积分。慢但是方便。

    RE : ndarray
        推迟势电场矢量REx, REy or REz
    t_ret : ndarray
        推迟时间矢量
    omega_axis : ndarray
        频谱的频率轴
    '''
    assert RE.ndim == 1 and t_ret.ndim == 1 and omega_axis.ndim == 1
    
    nomega = len(omega_axis)
    RE_ft = np.zeros(nomega, dtype=np.complex128)
    norm = 1 / np.sqrt(c*mu_0) / np.sqrt(2*pi)
    
    nan_mask = ~np.isnan(RE)
    RE_ = RE[nan_mask]
    t_ret_ = t_ret[nan_mask]
    nt_ = len(t_ret_)

    buf = np.zeros(nt_, dtype='c16')
    # definition of Fourier transformation
    for iw in range(nomega):
        w = omega_axis[iw]

        for it in range(nt_):
            buf[it] = RE_[it] * np.exp(1j*w*t_ret_[it])

        # trapzoid integral
        for it in range(nt_-1):
            dt = t_ret[it+1] - t_ret[it]
            RE_ft[iw] += 0.5*dt*(buf[it]+buf[it+1])
        RE_ft[iw] *= norm

    return RE_ft

@njit(parallel=True)
def get_RE_spectrum_2d_mt(RE, t_ret, omega_axis):
    '''
    计算lw谱，不使用FFT直接积分。慢但是方便。

    RE : ndarray
        推迟势电场矢量REx, REy or REz
    t_ret : ndarray
        推迟时间矢量
    omega_axis : ndarray
        频谱的频率轴
    '''

    nomega = len(omega_axis)
    npart, nt = RE.shape

    RE_ft = np.zeros((nomega, npart), dtype=np.complex128)
    norm = 1 / np.sqrt(c*mu_0) / np.sqrt(2*pi)
    
    # definition of Fourier transformation
    for ip in prange(npart):
        # trapzoid integral
        RE_ft[:, ip] = get_RE_spectrum_st(RE[ip], t_ret[ip], omega_axis)

    # normalize
    return RE_ft

get_RE_spectrum_2d_st = njit(get_RE_spectrum_2d_mt.py_func)


def get_lw_spectrum_cpu(
    x, y, z, ux, uy, uz, t, n, omega_axis, 
    mt=True,
):
    nomega = len(omega_axis)
    t_ret, REx, REy, REz = get_lw_RE(x, y, z, ux, uy, uz, t, n)
    I = np.zeros(nomega)
    for RE in (REx, REy, REz):
        if mt:
            RE_ft = get_RE_spectrum_mt(RE, t_ret, omega_axis)
        else:
            RE_ft = get_RE_spectrum_st(RE, t_ret, omega_axis)
        I += RE_ft.real**2 + RE_ft.imag**2
    I *= 2
    return I
    

def get_lw_spectrum_2d_cpu(
    x, y, z, ux, uy, uz, t, n, omega_axis, 
    mt=True,
):
    nt, npart = x.shape
    nomega = len(omega_axis)

    # buffer
    t_ret = np.zeros((npart, nt-2))
    REx = np.zeros((npart, nt-2))
    REy = np.zeros((npart, nt-2))
    REz = np.zeros((npart, nt-2))

    I = np.zeros((nomega, npart))

    for i in range(npart):
        t_ret[i, :], REx[i, :], REy[i, :], REz[i, :] = get_lw_RE(x[:, i], y[:, i], z[:, i], ux[:, i], uy[:, i], uz[:, i], t, n)
        
    for RE in (REx, REy, REz):
        if mt:
            RE_ft = get_RE_spectrum_2d_mt(RE, t_ret, omega_axis)
        else:
            RE_ft = get_RE_spectrum_2d_st(RE, t_ret, omega_axis)
        I += RE_ft.real**2 + RE_ft.imag**2
    I *= 2
    return I.sum(axis=1)

    



import unittest

import numpy as np
from scipy.constants import c, pi

from lwrad import get_lw_spectrum
from lwrad.cpu import get_lw_RE


class TestSpectrum(unittest.TestCase):
    l0 = 8e-7
    omega0 = 2*pi*c / l0

    def dummy_thomson(self):
        import h5py

        l0 = self.l0
        with h5py.File("tests/test.h5", "r") as f:
            x = f['x'][:, 0] / 2/pi * l0
            y = f['y'][:, 0] / 2/pi * l0
            z = f['z'][:, 0] / 2/pi * l0
            ux = f['px'][:, 0]
            uy = f['py'][:, 0]
            uz = f['pz'][:, 0]
            t = np.arange(len(ux)) * f.attrs['dt'] / 2/pi * l0/c

        return x, y, z, ux, uy, uz, t

    def test_thomson(self):
        omega0 = self.omega0
        nomega = 256
        gamma0 = 5.0

        omega_axis = np.linspace(0, 3, nomega) * 4*gamma0**2*omega0

        n = [1, 0, 0]
        
        I = get_lw_spectrum(*self.dummy_thomson(), n, omega_axis)
        self.assertAlmostEqual(I.max(), 1.7e-34, 2)


    def test_lw_RE(self):
        args = [np.concatenate(([np.nan]*10, x, [np.nan]*10) ) for x in self.dummy_thomson()]
        # test warn once
        with self.assertWarns(UserWarning):
            ret = get_lw_RE(*args, [1, 0, 0])
        
        for ret_ in ret:
            self.assertNotIn(np.nan, ret_)



import numpy as np
from scipy.constants import pi, m_e, e, c, alpha, hbar, epsilon_0, mu_0
import cupy as cp


from numba import njit, cuda, float64, complex128
import math, cmath
from scipy.constants import pi, epsilon_0, mu_0, c, e

@njit(inline="always")
def _calc_beta(ux, uy, uz):
    inv_gamma = 1 / math.sqrt(ux**2 + uy**2 + uz**2 + 1)
    betax = ux * inv_gamma
    betay = uy * inv_gamma
    betaz = uz * inv_gamma
    return betax, betay, betaz


@njit(inline="always")
def _calc_lw(
    t, x, y, z, 
    ux, uy, uz, 
    ux_prev, uy_prev, uz_prev, 
    ux_next, uy_next, uz_next, 
    nx, ny, nz, dt_next_prev
):
    betax, betay, betaz = _calc_beta(ux, uy, uz)
    betax_prev, betay_prev, betaz_prev = _calc_beta(ux_prev, uy_prev, uz_prev)
    betax_next, betay_next, betaz_next = _calc_beta(ux_next, uy_next, uz_next)

    # 加速度，假设首尾加速度为0
    ax = (betax_next - betax_prev) / dt_next_prev
    ay = (betay_next - betay_prev) / dt_next_prev
    az = (betaz_next - betaz_prev) / dt_next_prev

    n_dot_a = nx*ax + ny*ay + nz*az
    n_dot_beta = nx*betax + ny*betay + nz*betaz
    factor = 1 / (1 - n_dot_beta)**3
    factor *= e / (4*pi*epsilon_0*c)

    t_ret = t - (nx*x + ny*y + nz*z) / c
    REx = (n_dot_a*(nx - betax) + (n_dot_beta - 1) * ax) * factor
    REy = (n_dot_a*(ny - betay) + (n_dot_beta - 1) * ay) * factor
    REz = (n_dot_a*(nz - betaz) + (n_dot_beta - 1) * az) * factor
    return t_ret, REx, REy, REz


@cuda.jit
def kernel_lw_RE(
    t, x, y, z, ux, uy, uz, 
    nx, ny, nz,
    t_ret, REx, REy, REz,
):
    
    nt = len(t)
    it = cuda.grid(1)
    if it == 0 or it >= nt - 1:
        return
    
    t_ret[it-1], REx[it-1], REy[it-1], REz[it-1] = _calc_lw(
        t[it], x[it], y[it], z[it], 
        ux[it], uy[it], uz[it], 
        ux[it-1], uy[it-1], uz[it-1], 
        ux[it+1], uy[it+1], uz[it+1], 
        nx, ny, nz, t[it+1] - t[it-1]
    )
    
    
@cuda.jit
def kernel_lw_RE_2d(
    t, x, y, z, ux, uy, uz, 
    nx, ny, nz,
    t_ret, REx, REy, REz,
):
    
    nt = len(t)
    npart = x.shape[1]
    it, ip = cuda.grid(2)
    if it == 0 or it >= nt - 1 or ip >= npart:
        return
    
    t_ret[it-1, ip], REx[it-1, ip], REy[it-1, ip], REz[it-1, ip] = _calc_lw(
        t[it], x[it, ip], y[it, ip], z[it, ip], 
        ux[it, ip], uy[it, ip], uz[it, ip], 
        ux[it-1, ip], uy[it-1, ip], uz[it-1, ip], 
        ux[it+1, ip], uy[it+1, ip], uz[it+1, ip], 
        nx, ny, nz, t[it+1] - t[it-1]
    )

_TPB = _NBUF = 32
_NSHR = _TPB*3
@cuda.jit
def kernel_RE_spectrum(
    t_ret, REx, REy, REz,
    omega, 
    REx_ft_real, REy_ft_real, REz_ft_real,
    REx_ft_imag, REy_ft_imag, REz_ft_imag,
):
    
    nt = len(t_ret)
    nomega = len(omega)

    it, iomega = cuda.grid(2)
    tid = cuda.threadIdx.x

    buf = cuda.shared.array(_NSHR, complex128)
    buf[tid] = 0
    buf[tid+_NBUF] = 0
    buf[tid+2*_NBUF] = 0
    if it >= nt-1 or iomega >= nomega:
        return

    w = omega[iomega]

    next = it + 1
    if math.isnan(REx[it]) or math.isnan(REx[next]):
        return

    dt = t_ret[next] - t_ret[it]

    buf[tid]         = 0.5*dt*(REx[it]*cmath.exp(1j*w*t_ret[it]) + REx[next]*cmath.exp(1j*w*t_ret[next]))
    buf[tid+_NBUF]   = 0.5*dt*(REy[it]*cmath.exp(1j*w*t_ret[it]) + REy[next]*cmath.exp(1j*w*t_ret[next]))
    buf[tid+2*_NBUF] = 0.5*dt*(REz[it]*cmath.exp(1j*w*t_ret[it]) + REz[next]*cmath.exp(1j*w*t_ret[next]))

    cuda.syncwarp()
    s = 1
    while s < _NBUF:
        if tid % (2 * s) == 0:
            # Stride by `s` and add
            buf[tid] += buf[tid + s]
            buf[tid+_NBUF] += buf[tid+_NBUF + s]
            buf[tid+2*_NBUF] += buf[tid+2*_NBUF + s]
        s *= 2
        cuda.syncthreads()
    
    if tid == 0:
        norm = 1 / math.sqrt(c*mu_0) / math.sqrt(2*pi)

        REx_ft = buf[0] * norm
        REy_ft = buf[_NBUF] * norm
        REz_ft = buf[2*_NBUF] * norm
        cuda.atomic.add(REx_ft_real, iomega, REx_ft.real)
        cuda.atomic.add(REy_ft_real, iomega, REy_ft.real)
        cuda.atomic.add(REz_ft_real, iomega, REz_ft.real)
        cuda.atomic.add(REx_ft_imag, iomega, REx_ft.imag)
        cuda.atomic.add(REy_ft_imag, iomega, REy_ft.imag)
        cuda.atomic.add(REz_ft_imag, iomega, REz_ft.imag)


@cuda.jit
def kernel_RE_spectrum_2d(
    t_ret, REx, REy, REz,
    omega, 
    I
):
    nomega = len(omega)
    nt, npart = t_ret.shape

    iomega, ip = cuda.grid(2)

    if ip >= npart or iomega >= nomega:
        return

    w = omega[iomega]
    REx_ft = 0.0j
    REy_ft = 0.0j
    REz_ft = 0.0j
    for it in range(nt-1):
        if math.isnan(REx[it+1, ip]):
            continue

        dt = t_ret[it+1, ip] - t_ret[it, ip]
        # todo: check dt
        REx_ft += 0.5*dt*(REx[it, ip]*cmath.exp(1j*w*t_ret[it, ip])+REx[it+1, ip]*cmath.exp(1j*w*t_ret[it+1, ip]))
        REy_ft += 0.5*dt*(REy[it, ip]*cmath.exp(1j*w*t_ret[it, ip])+REy[it+1, ip]*cmath.exp(1j*w*t_ret[it+1, ip]))
        REz_ft += 0.5*dt*(REz[it, ip]*cmath.exp(1j*w*t_ret[it, ip])+REz[it+1, ip]*cmath.exp(1j*w*t_ret[it+1, ip]))


    norm = 1 / math.sqrt(c*mu_0) / math.sqrt(2*pi)

    REx_ft *= norm
    REy_ft *= norm
    REz_ft *= norm
    cuda.atomic.add(I, iomega, 2*(REz_ft.real**2 + REz_ft.imag**2 + REx_ft.real**2 + REx_ft.imag**2 + REy_ft.real**2 + REy_ft.imag**2))


def get_lw_spectrum_cuda(x, y, z, ux, uy, uz, t, n, omega_axis):
    nomega = len(omega_axis)
    REx_ft = cp.zeros(nomega, dtype='c16')
    REy_ft = cp.zeros(nomega, dtype='c16')
    REz_ft = cp.zeros(nomega, dtype='c16')

    nt = len(t)
    t_ret = cp.zeros(nt-2)
    REx = cp.zeros_like(t_ret)
    REy = cp.zeros_like(t_ret)
    REz = cp.zeros_like(t_ret)
    kernel_lw_RE[nt//_TPB+1, _TPB](
        t, x, y, z, ux, uy, uz, 
        n[0], n[1], n[2],
        t_ret, REx, REy, REz,
    )
    kernel_RE_spectrum[((nt-2)//_TPB + 1, nomega), (_TPB, 1)](
        t_ret, REx, REy, REz,
        cp.array(omega_axis),
        REx_ft.real, REy_ft.real, REz_ft.real,
        REx_ft.imag, REy_ft.imag, REz_ft.imag
    )
    I = 2*(REz_ft.real**2 + REz_ft.imag**2 + REx_ft.real**2 + REx_ft.imag**2 + REy_ft.real**2 + REy_ft.imag**2)
    return I.get()


def get_lw_spectrum_2d_cuda(x, y, z, ux, uy, uz, t, n, omega_axis, tpb=128):
    nomega = len(omega_axis)
    npart = x.shape[1]
    nt = len(t)
    
    t_ret = cp.zeros((nt-2, npart), order='F')
    REx = cp.zeros_like(t_ret)
    REy = cp.zeros_like(t_ret)
    REz = cp.zeros_like(t_ret)
    kernel_lw_RE_2d[(nt//tpb+1, npart), (tpb, 1)](
        t, x, y, z, ux, uy, uz,
        n[0], n[1], n[2],
        t_ret, REx, REy, REz,
    )

    I = cp.zeros(nomega)
    kernel_RE_spectrum_2d[(nomega, (nt-2)//tpb + 1), (1, tpb)](
        t_ret, REx, REy, REz,
        omega_axis,
        I
    )
    return I.get()



import numpy as np
from tqdm.auto import trange

from .cpu import c, get_lw_spectrum_2d_cpu, get_lw_spectrum_cpu


def _check_velocity(t, x, y, z):
    ndim = x.ndim

    dt = t[1:] - t[:-1]
    if ndim == 2:
        dt = dt[:, None]
    vx = (x[1:, ...] - x[:-1, ...]) / dt
    vy = (y[1:, ...] - y[:-1, ...]) / dt
    vz = (z[1:, ...] - z[:-1, ...]) / dt
    beta = np.sqrt(vx**2 + vy**2 + vz**2) / c
    if (beta >= 1).any():
        raise ValueError("speed greater than c, check input trajectories. \
                            set `check_velocity` to False to disable check.")
    

def get_lw_spectrum(
    x, y, z, ux, uy, uz, t, n, omega_axis, 
    backend='mt', check_velocity=True,
):
    '''
    从坐标和动量计算LW场的频谱

    Prameters
    ===
    x, y, z : ndarray
        坐标向量或矩阵。其中矩阵的行为时间、列为不同粒子
    ux, uy, uz : ndarray
        归一化动量，u = p/mc。结构与xyz相同。
    t : ndarray
        时间矢量
    n : tuple | list | ndarray
        长度为1的方向矢量
    omega_axis : ndarray
        频谱的频率轴
    backend : str
        mt | cuda | None
    check_velocity : bool
        是否检查xyz的轨迹是否超过光速
    
    Returns
    ===
    I: ndarray
        返回dI/dΩdω
    '''

    assert backend in ['mt', 'cuda', None]
    assert omega_axis.ndim == 1
    
    if backend=='cuda':
        try:
            from .cuda import get_lw_spectrum_2d_cuda, get_lw_spectrum_cuda
        except ImportError as e:
            raise e
        
    if check_velocity:
        _check_velocity(t, x, y, z)
    
    args = (x, y, z, ux, uy, uz, t, n, omega_axis)
    argsT = (x.T, y.T, z.T, ux.T, uy.T, uz.T, t, n, omega_axis)
    if x.ndim == 1:
        if backend=='cuda':
            return get_lw_spectrum_cuda(*args)
        if backend == 'mt':
            return get_lw_spectrum_cpu(*args, mt=True)
        if backend is None:
            return get_lw_spectrum_cpu(*args, mt=False)
        
    elif x.ndim == 2:
        if x.shape[0] == x.shape[1]:
            print("input trajectory is a square matrix, assume the 2nd axis is time.")
        
        if len(t) == x.shape[1]:
            if backend=='cuda':
                return get_lw_spectrum_2d_cuda(*argsT)
            if backend == 'mt':
                return get_lw_spectrum_2d_cpu(*argsT, mt=True)
            if backend is None:
                return get_lw_spectrum_2d_cpu(*argsT, mt=False)
        elif len(t) == x.shape[0]:
            if backend=='cuda':
                return get_lw_spectrum_2d_cuda(*args)
            if backend == 'mt':
                return get_lw_spectrum_2d_cpu(*args, mt=True)
            if backend is None:
                return get_lw_spectrum_2d_cpu(*args, mt=False)
        else:
            raise TypeError("length of t does not match input 2d trajectory.")
    else:
        raise TypeError("input x, y, z, ux, uy, uz must 1d vector or 2d array.")

        
def get_lw_angular_spectrum(
    x, y, z, ux, uy, uz, t, omega_axis,
    theta_max, ntheta, theta_min=0.0,
    direction="x", theta_plane="xy",
    backend='mt', check_velocity=True,
    comm=None,
):
    '''
    从坐标和动量计算LW场的频谱

    Prameters
    ===
    x, y, z : ndarray
        坐标向量或矩阵。其中矩阵的行为时间、列为不同粒子
    ux, uy, uz : ndarray
        归一化动量，u = p/mc。结构与xyz相同。
    t : ndarray
        时间矢量
    omega_axis : ndarray
        频谱的频率轴
    theta_max, theta_min : float
        角度范围参数
    ntheta : int
        角度数量
    direction : "x", "y" or "z"
        扫描角度的中心方向
    theta_plane : "xy", "yz" or "xz"
        角度范围所在的平面
    use_cuda : bool
        是否用gpu计算频谱，需要安装cupy
    check_velocity : bool
        是否检查xyz的轨迹是否超过光速
    
    Returns
    ===
    I: ndarray
        返回dI/dΩdω
    '''
    assert direction in ["x", "y", "z"]
    if direction == "x":
        assert theta_plane in ["xy", "xz"]
        if theta_plane == "xy":
            n = lambda theta : [np.cos(theta), np.sin(theta), 0.0]
        if theta_plane == "xz":
            n = lambda theta : [np.cos(theta), 0.0, np.sin(theta)]
    if direction == "y":
        assert theta_plane in ["xy", "yz"]
        if theta_plane == "xy":
            n = lambda theta : [np.sin(theta), np.cos(theta), 0.0]
        if theta_plane == "yz":
            n = lambda theta : [0.0, np.cos(theta), np.sin(theta)]
    if direction == "z":
        assert theta_plane in ["xz", "yz"]
        if theta_plane == "xz":
            n = lambda theta : [np.sin(theta), 0.0, np.cos(theta)]
        if theta_plane == "yz":
            n = lambda theta : [0.0, np.sin(theta), np.cos(theta)]
        

    nomega = len(omega_axis)
    theta_axis = np.linspace(theta_min, theta_max, ntheta)
    spectrum = np.zeros((ntheta, nomega))

    # TODO: 
    # 多个角度调用2d函数,
    # jit函数分段sum
    
    if comm:
        rank = comm.Get_rank()
        comm_size = comm.Get_size()
    else:
        rank = 0
        comm_size = 1
        
    for itheta in trange(rank, ntheta, comm_size, position=rank):
        theta = theta_axis[itheta]
        spectrum[itheta, :] += get_lw_spectrum(x, y, z, ux, uy, uz, t, n(theta), omega_axis, backend, check_velocity)
    
    if comm_size > 1:
        spectrum = comm.reduce(spectrum)
        
    return theta_axis, spectrum


from pathlib import Path

import h5py
import numpy as np
import typer
from typing_extensions import Annotated

from .utils import get_lw_angular_spectrum


def generic(path: Path):
    """
    Generic particle file.
    """
    args = []
    with h5py.File(path, "r") as f:
        for prop in  ['x', 'y', 'z', 'ux', 'uy', 'uz', 't']:
            args.append(f[prop][()])
            
    return args

def smilei(path: Path, name: str):
    """
    Generate a LW angular spectrum from a SMILEI simulation and save it to the specified file.
    """
    track_particle_disordered_file = path / f"TrackParticlesDisordered_{name}.h5"
    track_particle_file = path / f"TrackParticles_{name}.h5"
    
            
    if not track_particle_disordered_file.exists():
        raise FileNotFoundError(f"{track_particle_disordered_file} not found.")
    
    if not track_particle_file.exists():
        print(f"{track_particle_file} not found. maybe the particles are not sorted by happi.")
        print(f"try sorting with happi...")
        try:
            import happi
            sim = happi.Open(path)
            sim.TrackParticles(name, sort=True)
        except ImportError:
            raise Exception("happi is required to sort the particle files")
        except Exception as e:
            raise e

    with h5py.File(track_particle_disordered_file, "r") as f:
        ts = list(f['data'].keys())
        dtSI = f['data'][ts[0]].attrs['dt'] * f['data'][ts[0]].attrs['timeUnitSI']
        dxSI = f['data'][ts[0]]['particles'][name]['position']['x'].attrs['unitSI']
    
    
    with h5py.File(track_particle_file, "r") as f:
        x = f['x'] * dxSI
        
        args = [x, ]
        for prop in ['y', 'z']:
            try:
                args.append(f[prop] * dxSI)
            except Exception:
                args.append(np.zeros_like(x))
        for prop in ['px', 'py', 'pz']:
            try:
                args.append(f[prop][()])
            except Exception:
                args.append(np.zeros_like(x))
        
        t = f['Times']*dtSI
        args.append(t)
    print(f"shape of input trajectory: x.shape={x.shape}, t.shape={t.shape}")
        
    return args
        
def epoch(path: Path, name: str):
    raise NotImplementedError()

def main(
    path: Annotated[Path, typer.Argument()], 
    from_smilei: Annotated[bool, typer.Option("--smilei")]=False,
    from_epoch: Annotated[bool, typer.Option("--epoch")]=False,
    name: Annotated[str, typer.Argument(help='species or particle name, required for Smilei and EPOCH files')]=None,
    ek_max: Annotated[float, typer.Option(help="Maximum energy of axis [MeV]")]=1.0,
    nek: Annotated[int, typer.Option()]=128,
    theta_max: Annotated[float, typer.Option()]=0.5,
    ntheta: Annotated[int, typer.Option()]=128,
    direction: Annotated[str, typer.Option(help='x | y | z')]='x',
    theta_plane: Annotated[str, typer.Option(help='xy | xz | yz')]='xy',
    backend: Annotated[str, typer.Option(help="mt | cuda | None")]='mt',
    check_velocity: Annotated[bool, typer.Option()]=True,
    savefig: Annotated[bool, typer.Option()]=False,
):
    """
    Calculate angular radiation spectrum from relativistic particle trajectories using Lienard-Wiechert potentials.
    """
    if from_smilei:
        assert name is not None, "Name of particle must be specified for Smilei files."
        args = smilei(path, name)
    elif from_epoch:
        assert name is not None, "Name of particle must be specified for Epoch files."
        args = epoch(path, name)
    else:
        args = generic(path)
        
    try:
        from mpi4py.MPI import COMM_WORLD as comm
        rank = comm.Get_rank()
    except ImportError:
        comm = None
        rank = 0
        
    if backend == 'None':
        backend = None
        
    ek_axis = np.linspace(0, ek_max, nek)
    omega_axis = ek_axis*1e6/1.55*2.35e15
    theta_axis, spectrum = get_lw_angular_spectrum(
        *args, 
        omega_axis=omega_axis, 
        theta_max=theta_max, 
        ntheta=ntheta, 
        direction=direction, 
        theta_plane=theta_plane, 
        backend=backend, 
        check_velocity=check_velocity,
        comm=comm
    )
    if rank == 0:
        savepath = path if path.is_dir() else path.parent
        np.savetxt(savepath/'theta_axis.txt', theta_axis)
        np.savetxt(savepath/'spectrum.txt', spectrum)
        if savefig:
            try:
                import matplotlib.pyplot as plt
            except ImportError as e:
                raise e
            fig, ax = plt.subplots(layout='tight')
            h = ax.imshow(
                spectrum,
                extent = [ek_axis[0], ek_axis[-1], theta_axis[0], theta_axis[-1]],
                aspect='auto',
                origin='lower'
            )
            fig.colorbar(h, ax=ax)
            fig.savefig(savepath/'spectrum.png', dpi=300)